from typing import Optional
import click
import logging
import enlighten
import time
from .schemas import SchemaBuilder
from .schedulers import Scheduler
from . import __version__


@click.command()
def demo():
    """Try a schema scheduling demo."""
    click.echo("""
Please give a schema file:

    $ python -m schemdule --schema schema.py
""")

    demo_schema = """
from datetime import datetime, timedelta
now = datetime.now()

def callable_payload():
    print("From Callable")

at((now + timedelta(seconds=1)).time(), "Demo event", callable_payload)

at((now + timedelta(seconds=3)).time(), "Demo event")

cycle((now + timedelta(seconds=5)).time(), 
    (now + timedelta(seconds=20)).time(),
    "00:00:05",
    "00:00:05", "Demo cycle")

prompter.clear().useSwitcher().useConsole().useCallable()
"""

    click.echo("A demo schema:\n")
    click.echo("\n".join(map(lambda x: "    " + x,
                             demo_schema.strip().splitlines())))

    click.echo("\nScheduling...")

    tt = SchemaBuilder()
    sc = Scheduler()

    tt.load(demo_schema)

    sc.schedule(tt.result)


@click.command()
@click.option("--preview", is_flag=True, default=False, help="Preview the built timetable.")
@click.argument("schema", type=click.Path(exists=True))
def run(schema: str, preview: bool = False) -> None:
    """Schedule a schema."""
    logger = logging.getLogger("run")

    click.echo(f"Welcome to Schemdule v{__version__}!")

    tt = SchemaBuilder()

    with open(schema, encoding="utf8") as f:
        tt.load(f.read())

    if preview:
        for item in sorted(tt.result.items):
            click.echo(f"{item.message} @ {item.time}")
    else:
        sc = Scheduler()
        sc.schedule(tt.result)


@click.command()
def ext() -> None:
    """List all installed extensions."""
    logger = logging.getLogger("ext")

    from .extensions import find_extensions, load_extension, get_extension_metadata

    extnames = find_extensions()

    click.echo(f"Found {len(extnames)} extension(s).")

    for name in extnames:
        ext = load_extension(name)
        metadata = get_extension_metadata(ext)
        click.echo(f"  {name} v{metadata['version']}")


@click.group(invoke_without_command=True)
@click.pass_context
@click.option('-v', '--verbose', count=True, default=0, type=click.IntRange(0, 4))
@click.option("--version", is_flag=True, default=False, help="Show the version.")
def main(ctx, verbose: int = 0, version: bool = False) -> None:
    """Schemdule (https://github.com/StardustDL/schemdule)

A tiny tool using script as schema to schedule one day and remind you to do something during a day.
"""
    logger = logging.getLogger("Cli-Main")

    loggingLevel = {
        0: logging.ERROR,
        1: logging.WARNING,
        2: logging.INFO,
        3: logging.DEBUG,
        4: logging.NOTSET
    }[verbose]

    logging.basicConfig(level=loggingLevel)

    logger.debug(f"Logging level: {loggingLevel}")

    if version:
        click.echo(f"Schemdule v{__version__}")
        exit(0)


main.add_command(run)
main.add_command(demo)
main.add_command(ext)

if __name__ == '__main__':
    main()



from enum import auto
from typing import Any
import click

from . import Prompter, PromptResult


class TkinterMessageBoxPrompter(Prompter):
    def __init__(self, final: bool = False) -> None:
        super().__init__(final)

    def prompt(self, message: str, payload: Any) -> Any:
        import tkinter
        import tkinter.messagebox

        top = tkinter.Tk()
        top.withdraw()
        tkinter.messagebox.showinfo(f"Attention {message}", payload)
        top.destroy()

        return self.success()


class ConsolePrompter(Prompter):
    def __init__(self, final: bool = False) -> None:
        super().__init__(final)

    def prompt(self, message: str, payload: Any) -> Any:
        click.echo(f"Attention {message}: {payload}")

        return self.success()


class CallablePrompter(Prompter):
    def __init__(self, final: bool = False) -> None:
        super().__init__(final)

    def prompt(self, message: str, payload: Any) -> Any:
        if callable(payload):
            result = payload()
            if isinstance(result, PromptResult):
                return result
            return self.success()
        return PromptResult.Unsupported



from abc import ABC, abstractmethod
import logging
from types import ModuleType
from typing import Any, Dict, List, Set
import importlib
import pkgutil

EXTENSION_MODULE_PREFIX = "schemdule.extensions."

_logger = logging.getLogger("extensions")


def schemaEnvironment(env: Dict[str, Any]) -> None:
    pass


def find_extensions() -> List[str]:
    return [name.replace(EXTENSION_MODULE_PREFIX, "", 1) for finder, name, ispkg in pkgutil.iter_modules(__path__, EXTENSION_MODULE_PREFIX)]


def import_extension(full_name: str) -> ModuleType:
    _logger.debug(f"Import extension {full_name}.")
    return importlib.import_module(f"{full_name}.__reg__")


def load_extension(name: str) -> ModuleType:
    _logger.info(f"Load extension {name}.")
    return import_extension(f"schemdule.extensions.{name}")


def load_extensions(names: List[str]) -> List[ModuleType]:
    return [load_extension(name) for name in names]


def use_extension(extension: ModuleType, env: Dict[str, Any]) -> None:
    _logger.info(
        f"Use extension {extension.__name__.replace('__reg__', '', 1)}.")
    configFuncName = "schemaEnvironment"
    if hasattr(extension, configFuncName):
        func = getattr(extension, configFuncName)
        func(env)


def use_extensions(extensions: List[ModuleType], env: Dict[str, Any]) -> None:
    for extension in extensions:
        use_extension(extension, env)


def get_extension_metadata(extension: ModuleType) -> None:
    version = getattr(extension, "__version__", None)
    return {"version": version}



from abc import ABC, abstractmethod
from typing import Any, Iterable, List, Optional
from enum import Enum


class PromptResult(Enum):
    Empty = 0
    Unsupported = 1     # Unsupported, skip this and go next
    Resolved = 2        # Resolved, go next
    Finished = 3        # Resolved, stop going next
    Failed = 4          # Failed, stop going next


class Prompter(ABC):
    def __init__(self, final: bool = False) -> None:
        self.final = final

    def success(self) -> PromptResult:
        return PromptResult.Finished if self.final else PromptResult.Resolved

    @abstractmethod
    def prompt(self, message: str, payload: Any) -> Any:
        pass

    def __repr__(self) -> str:
        return type(self).__name__


class PrompterHub(Prompter, ABC):
    @abstractmethod
    def register(self, prompter: Prompter) -> None:
        pass


class PrompterPayload(ABC):
    pass


class PrompterPayloadCollection(PrompterPayload):
    def __init__(self, payloads: Optional[List[Any]]) -> None:
        self.payloads = payloads if payloads is not None else []

    def try_get(self, type: type) -> Iterable[Any]:
        for item in self.payloads:
            if isinstance(item, type):
                yield item



from datetime import date, time, datetime, timedelta
from typing import Optional, Union, Any

import functools
from queue import deque
import enlighten
import click
import json
import logging

from time import sleep

from ..prompters import Prompter
from ..prompters.configer import PrompterConfiger
from ..extensions import load_extension, use_extension
from ..timeutils import to_timedelta, subtract_time, parse_time


@functools.total_ordering
class TimeTableItem:
    def __init__(self, time: time, message: str = "", payload: Any = None) -> None:
        self.time = time
        self.message = message
        self.payload = payload

    def __eq__(self, other: object) -> bool:
        if isinstance(other, TimeTableItem):
            return (self.time, self.message) == (other.time, other.message)
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, TimeTableItem):
            return self.time < other.time
        return NotImplemented

    def __repr__(self) -> str:
        return f"TimeTableItem({self.time}, {self.message})"


class TimeTable:
    _logger = logging.getLogger("TimeTable")

    def __init__(self) -> None:
        self.items: list[TimeTableItem] = []
        self.prompter: Optional[Prompter] = None

    def use(self, prompter: Optional[Prompter]) -> None:
        self._logger.debug(f"Use prompter {prompter}.")
        self.prompter = prompter

    def at(self, time: time, message: str = "", payload: Any = None) -> None:
        self._logger.debug(f"{message} ({payload}) at {time}.")
        self.items.append(TimeTableItem(time, message, payload))

    def cycle(self, start: time, end: time, work_duration: time, rest_duration: time, message: str = "", payload: Any = None) -> None:
        self._logger.debug(
            f"{message} ({payload}) cycle from {start} to {end} (work {work_duration}, rest {rest_duration}).")
        _start = datetime(2000, 1, 1) + to_timedelta(start)
        _end = datetime(2000, 1, 1) + to_timedelta(end)
        _work_duration = to_timedelta(work_duration)
        _rest_duration = to_timedelta(rest_duration)

        index = 0

        current = _start

        while current < _end:
            index += 1
            self.at(min(current, _end).time(),
                    f"{message} (cycle {index} starting)", payload)
            current += _work_duration
            self.at(min(current, _end).time(),
                    f"{message} (cycle {index} resting starting)", payload)
            current += _rest_duration

    def clear(self) -> None:
        self.items.clear()



from typing import Optional


from . import Prompter, PrompterHub
from .hubs import PrompterBroadcaster, PrompterSwitcher
from .general import ConsolePrompter, TkinterMessageBoxPrompter, CallablePrompter


class PrompterConfiger:
    def __init__(self) -> None:
        self._result: Optional[Prompter] = None

    def use(self, prompter: Prompter) -> "PrompterConfiger":
        if self._result is None:
            self._result = prompter
        elif isinstance(self._result, PrompterHub):
            self._result.register(prompter)
        else:
            raise Exception(
                f"The prompter is set to a no-hub prompter {type(self._result)}.")
        return self

    def useBroadcaster(self, final: bool = False) -> "PrompterConfiger":
        return self.use(PrompterBroadcaster(final))

    def useSwitcher(self, final: bool = False) -> "PrompterConfiger":
        return self.use(PrompterSwitcher(final))

    def useConsole(self, final: bool = False) -> "PrompterConfiger":
        return self.use(ConsolePrompter(final))

    def useCallable(self, final: bool = False) -> "PrompterConfiger":
        return self.use(CallablePrompter(final))

    def useTkinterMessageBox(self, final: bool = False) -> "PrompterConfiger":
        return self.use(TkinterMessageBoxPrompter(final))

    def clear(self) -> "PrompterConfiger":
        self._result = None
        return self

    def build(self) -> Prompter:
        return self._result



from datetime import datetime, time, timedelta

BASE_DATETIME = datetime(2000, 1, 1)


def to_timedelta(value: time) -> timedelta:
    return timedelta(hours=value.hour, minutes=value.minute,
                     seconds=value.second, microseconds=value.microsecond)


def subtract_time(a: time, b: time) -> timedelta:
    return to_timedelta(a) - to_timedelta(b)


def parse_time(value: str) -> time:
    return time(*list(map(int, map(round, map(float, value.split(':'))))))



from datetime import date, time, datetime, timedelta
from typing import Dict, Optional, Union, Any

import functools
from queue import deque
import enlighten
import click
import json
import logging

from ..prompters import Prompter
from ..prompters.configer import PrompterConfiger
from ..extensions import load_extension, use_extension, find_extensions, load_extensions, use_extensions
from ..timeutils import to_timedelta, subtract_time, parse_time
from .timetable import TimeTable, TimeTableItem


def default_prompter_configer() -> PrompterConfiger:
    prompter = PrompterConfiger()
    prompter.useSwitcher().useConsole().useCallable(True).useTkinterMessageBox()
    return prompter


class SchemaBuilder:
    _logger = logging.getLogger("SchemaBuilder")

    def __init__(self) -> None:
        self.result: TimeTable = TimeTable()

    def get_env(self) -> Dict[str, Any]:
        prompterConfiger = default_prompter_configer()

        env = {}

        def at(raw_time: Union[str, time], message: str = "", payload: Any = None):
            ttime = parse_time(raw_time) if isinstance(
                raw_time, str) else raw_time
            self.result.at(ttime, message, payload)

        def cycle(raw_start: Union[str, time], raw_end: Union[str, time], raw_work_duration: Union[str, time], raw_rest_duration: Union[str, time], message: str = "", payload: Any = None):
            tstart = parse_time(raw_start) if isinstance(
                raw_start, str) else raw_start
            tend = parse_time(raw_end) if isinstance(raw_end, str) else raw_end
            twork_duration = parse_time(raw_work_duration) if isinstance(
                raw_work_duration, str) else raw_work_duration
            trest_duration = parse_time(raw_rest_duration) if isinstance(
                raw_rest_duration, str) else raw_rest_duration
            self.result.cycle(
                tstart, tend, twork_duration, trest_duration,
                message, payload)

        def load(source: str):
            src_preview = source[:50].replace('\n', ' ').replace('\r', ' ')
            self._logger.info(f"Load: '{src_preview}...'")
            exec(source, env)

        def ext(name: Optional[str] = None):
            if name is None:
                self._logger.info("Use all installed extensions.")
                extnames = find_extensions()
                exts = load_extensions(extnames)
                use_extensions(exts, env)
            else:
                self._logger.info(f"Use extension {name}.")
                extension = load_extension(name)
                use_extension(extension, env)

        env["at"] = at
        env["cycle"] = cycle
        env["load"] = load
        env["ext"] = ext
        env["prompter"] = prompterConfiger
        env["env"] = env

        return env

    def use_env(self, env: Dict[str, Any]):
        prompter = env.get("prompter")
        if isinstance(prompter, PrompterConfiger):
            self.result.use(prompter.build())

    def load_with_env(self, src: str, env: Dict[str, Any]) -> None:
        env["load"](src)
        self.use_env(env)

    def load(self, src) -> None:
        self.load_with_env(src, self.get_env())



from datetime import date, time, datetime, timedelta
from typing import Optional, Union, Any

import functools
from queue import deque
import enlighten
import click
import json
import logging

from time import sleep
from ..prompters import Prompter
from ..schemas.timetable import TimeTable, TimeTableItem
from ..schemas import default_prompter_configer
from ..timeutils import subtract_time


class Scheduler:
    _logger = logging.getLogger("Scheduler")

    def schedule(self, timetable: TimeTable, prompter: Optional[Prompter] = None) -> None:
        def outdating(item: TimeTableItem) -> bool:
            now = datetime.now().time()

            if item.time < now:
                self._logger.info(f"Outdated: {item}.")
                click.echo(f"Outdated: {item.message} @ {item.time}")
                return True

            return False

        def pending(item: TimeTableItem, status: enlighten.StatusBar, manager: enlighten.Manager) -> bool:
            now = datetime.now().time()

            if item.time <= now:
                return False

            self._logger.info(f"Pending: {item}.")
            status.update(f"Pending: {item.message} @ {item.time}")
            deltaNow = subtract_time(item.time, now)
            pendingTotal = int(round(deltaNow.total_seconds()))

            with manager.counter(
                    total=pendingTotal, desc="", unit='ticks', leave=False) as pbar:

                while True:
                    now = datetime.now().time()
                    if item.time <= now:
                        self._logger.info(f"Occurring: {item}.")
                        click.echo(f"Attention: {item.message} @ {item.time}")
                        result = prompter.prompt(item.message, item.payload)
                        self._logger.info(f"Prompting result: {result}.")
                        return True
                    else:
                        delta = subtract_time(item.time, now)
                        count = int(round(delta.total_seconds()))
                        pbar.update((pbar.total - count) - pbar.count)
                    sleep(1)

            return False

        self._logger.info(f"Start scheduling.")
        click.echo(f"Started Time: {datetime.now().time()}")

        prompter = timetable.prompter if prompter is None else prompter

        if prompter is None:
            prompter = default_prompter_configer().build()

        self._logger.info(f"Used prompter: {prompter}.")

        items = deque(sorted(timetable.items))

        with enlighten.get_manager() as manager:
            with manager.status_bar('Status',
                                    color='white_on_cyan',
                                    justify=enlighten.Justify.CENTER, leave=False) as status:

                while len(items) > 0:
                    item: TimeTableItem = items[0]

                    if outdating(item):
                        items.popleft()
                    elif pending(item, status, manager):
                        items.popleft()



import logging
from typing import Any
from . import Prompter, PromptResult, PrompterHub


class PrompterSwitcher(PrompterHub):
    _logger = logging.getLogger("PrompterSwitcher")

    def __init__(self, final: bool = False) -> None:
        super().__init__(final)
        self.prompters: list[Prompter] = []

    def register(self, prompter: Prompter) -> None:
        self._logger.debug(f"Register: {prompter}.")
        self.prompters.append(prompter)

    def prompt(self, message: str, payload: Any) -> Any:
        results = []

        for prompter in self.prompters:
            self._logger.info(f"Send '{message}' to prompter: {prompter}.")

            result = prompter.prompt(message, payload)

            self._logger.info(f"Recieved {result} from prompter {prompter}.")

            if not isinstance(result, PromptResult):
                result = PromptResult.Empty

            results.append(result)

            if result is PromptResult.Finished:
                self._logger.info(f"Finish prompting for '{message}'.")
                break
            elif result is PromptResult.Failed:
                self._logger.error(
                    f"Failed prompting for '{message}' at prompter {prompter}.")
                return PromptResult.Failed

        lr = len(results)

        if lr == 0 or len([x for x in results if x is PromptResult.Empty]) == lr:
            self._logger.info(f"Empty prompting for '{message}'.")
            return PromptResult.Empty
        elif len([x for x in results if x is PromptResult.Unsupported]) == lr:
            self._logger.info(f"Unsupported prompting for '{message}'.")
            return PromptResult.Unsupported
        else:
            self._logger.info(f"Success prompting for '{message}'.")
            return self.success()


class PrompterBroadcaster(PrompterHub):
    _logger = logging.getLogger("PrompterBroadcaster")

    def __init__(self, final: bool = False) -> None:
        super().__init__(final)
        self.prompters: list[Prompter] = []

    def register(self, prompter: Prompter) -> None:
        self._logger.debug(f"Register: {prompter}.")
        self.prompters.append(prompter)

    def prompt(self, message: str, payload: Any) -> Any:
        for prompter in self.prompters:
            self._logger.info(f"Send '{message}' to prompter: {prompter}.")

            result = prompter.prompt(message, payload)

            self._logger.info(f"Recieved {result} from prompter {prompter}.")

            if result is PromptResult.Failed:
                self._logger.error(
                    f"Failed prompting for '{message}' at prompter {prompter}.")
                return PromptResult.Failed

        self._logger.info(f"Success prompting for '{message}'.")
        return self.success()
